%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Server Class of the fault tolerant transactional model
% Author:      Ilies Alouini, Mostafa Al-Metwally
% @copyright UCL & Region Wallone              
% Read LICENSE                                 
% Last update: 15-01-02 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Class Server: Implements the deduction rules described in
% the paper: Rules  R14 R15 R16 R17 
% Reflective failure handling for collaborative applications 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
functor
import
   System
   Fault
   Application
export
  ServerClass
define
   %%%%%%%%%%%%%%%%%%%%%%%
   %% Debug Show procedure
   %%%%%%%%%%%%%%%%%%%%%%%
   Debug=0
   proc {Show Message}
      if Debug==1
      then
	 {System.show Message}
      else
	 skip
      end
   end
 
   %%%%%%%%%%%%%%%%%%%%%%%
   %% Server Class
   %%%%%%%%%%%%%%%%%%%%%%%   
   class ServerClass
      prop locking
      attr
	 locks           %Dictionary: pairs of (objectidentifier,stationaryclientobject)
	 clientset:nil   %Client stationary object set
	 identset:nil    %Client ident set added by metwally
	 identclient:nil %Clientref#clientident
	 server:noserver %Server stationary object	
	 procwatcherclient:proc{$ E C} skip end
	 idclient:0      %Counter to generate a unique identifier to each connected client
	 idobject:0      %Counter to generate a unique identifier for each object
	 listconnectedclient:nil % used for connection protocol
	 oldserver       % server with tempFail
	 terminate:0     % to detect termination
	 listobjects:nil % list of broadcastedobjects
	 membership:nil  %list of Username#id
	 listofwatcher: nil% watcher installed on clients
	 requestlock:true
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Init the Server Class
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
      meth init(Server)
	 locks<-{NewDictionary $}
	 server <- Server
      end   
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      % Rule R14: Get all_locks
      % Rule R15: Get no_locks
      % Rule R22: Recovery all_locks message
      % Rule R23: Recovery no_locks message
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      meth request_locks(Sel Transid ClientId)
	 Client Id in
	 lock
	   ClientId=Client#Id
	    terminate <- 0
	    {Show 'server:request_locks:begin:'#Transid#Id}
	    try
	      
	       {ForAll Sel  % We should test if the object in the store or not (remember deleted obects  listobjects( broadcasting object)
		proc {$ Objid }		
		   if {Dictionary.condGet @locks Objid noentry $}==noentry
		   then
		      skip
		   else  
		      raise nolock end
		   end
		end
	       }
               % all  objects in selection Sel are locked by Client    
	       {ForAll Sel
		proc {$ Objid }
		   {Dictionary.put @locks Objid Client}
		end
	       }
	       % send all_locks(id) to C
	       try
		  {Client serv_all_locks(Transid Sel)}
	       catch senderror then
		  % Rule 22
		  skip			     
	       end
	    catch nolock
	    then
	       try
		  {Client serv_no_locks(Transid Sel)}
	       catch senderror then
		  % Rule 23
		  skip			     
	       end
	    end
	    {Show 'server:request_locks:end:'#Transid#Id}
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      % Synchronous request of locks 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth waitlocks(Transid Sel Res Client )
	 lock
	    terminate <- 0
	    try
	       {ForAll Sel 
		proc {$ Objid }		
		   if {Dictionary.condGet @locks Objid noentry $}==noentry
		   then
		      skip
		   else  
		      raise nolock end
		   end
		end
	       }
	       
	       {ForAll Sel
		proc {$ Objid }
		   {Dictionary.put @locks Objid Client}
		end
	       }
	      
	       try Res=true
	       catch _  % failure when binding
	       then skip
	       end
	       
	    catch nolock then
	       try Res=false
	       catch _  % failure when binding
	       then skip
	       end
	    end
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      % Rule R13 release locks
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      meth release_locks(Sel Transid C)
	 lock
	    {Show 'server:release_locks:'#Transid#C}
	    {ForAll Sel
	     proc {$ Objid}
		{Dictionary.remove @locks Objid}
	     end
	    }
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      % Rule R14: broadcast updates
      % Rule R24: Recovery update message
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth update(Methods Transid C)	 
	 lock
	    terminate <- 0
	    if Methods==nil then
	       skip
	       %{System.show  'INTERNAL_ERROR'} {Application.exit 0}
	    else
	       {ForAll @clientset
		proc {$ Client}		
		   if Client\=C then
		      try
			 % test if updates already broadcasted by connection
			 % protocol.
			 % it is sufficient to test the last transaction id
			 try
			    {Show 'server:listconnectedclient:'#@listconnectedclient} 
			    {ForAll @listconnectedclient
			     proc {$ X}
				Trid Cl  in
				X=Trid#Cl
				%listconnectedclient <- {List.subtract @listconnectedclient X $}
				if Trid \= none then
				   if Cl==C andthen Trid>=Transid
				   then
                                   % new connected client
				   % don't send updates
				      
				      %{System.show 'server:update:listconnectedclient:nosend'#Trid#Transid}
				      
				      raise nosend end
				   else
				      raise send end
				   end
				else
				   skip
				end
			     end
			    }
			    raise send end
			 catch
			    send
			 then
			    {Client update(Methods)}
			 [] nosend then {Show nosend} skip
			 end
		      catch senderror then
			 skip
		      end 
		   else
		      skip
		   end
		end
	       }	     	       	       
	    end
	 end
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      % R17 Client failure
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth procwatcher(Client)
	 lock
	    %{System.show 'Client:failed'}
	    {@server disconnectstore(Client)}
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Methods to handle tempfail
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
      meth initiate_merge(Oldserver)
	 X in
         % send a ping message to the old server
         % this call will blocks until tempfail diseappears
	 % Change Faultstate on server
	 %{Fault.enable thread permfail....sur server}
	 thread
	    oldserver <- Oldserver
	    {Oldserver ping(X)}
	    {Wait X}
            % Old server alive: do the merge between (Oldserver,self)
	    {@server do_merge1(Oldserver)}
	 end
      end
      meth do_merge1(Oldserver) 
         % Forall clients initiate an election
	 {ForAll @clientset
	  proc {$ C}
	     {C watcheronserver(nofail)}
	  end
	  }	 
      end
      meth store(Unlockedtore Lockedstore)
        % merge the store with the old server
	 {Show 'store:'#Unlockedtore#Lockedstore}
        %{@oldserver do_merge2(Unlockedtore Lockedstore @clientset)}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Init store at a connected client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth connectstore(Client Username ?UserId ?Members ?Store )
	 lock
	    Lastcommitransid
	    proc {ProcT E FS}
	       {Show 'temporary fail detected'}
	    end
	    proc{Proc E FS}
	       {self procwatcher(Client)}
	    end
	    
	    Synch
	 in
	    {Show 'server:initconnectstore:'#Client}
	    % set  the server attribute
	    try
	       {Client installwatcher([permFail] Proc Fault)}
	       {Show 'server:watcherinstalleokclient:'#Client}
	       %{Client installtempwatcher(ProcT Fault)}
	       procwatcherclient <- Proc
	       listofwatcher <- Client#Proc|@listofwatcher
	    catch cannotinstallwatcher
	    then
	       {Show installwatchererror}
	    end
	    % choose one unfailed client and gets the store
	    Members=@membership
	    UserId=@idclient+1
	    try
	       {ForAll @clientset
		proc {$ Item}
		  
		if Item\=Client then 
		   try	      
		     
		     % Store and Lastcommitransid will be bound at client
		      {Item getstore(?Store ?Lastcommitransid ?Synch)} %
		      		     
		      {Wait Synch}
		      % used to get coherent store
	              listconnectedclient <- Lastcommitransid#Item|@listconnectedclient
		      raise endforall end
		   catch
		      senderror then skip
		   [] endforall then raise endforall end
		   end
		   else
		      skip
		   end
	     end
	    }
	    catch endforall then
	    skip
	    end
	      
	    % set the global store to the connected client
	    if {List.length @clientset}==0 then
	       
	       Store=nil#0
	       Lastcommitransid=none 
	    else
	       skip
	    end	    
	    {Show 'server: getstore:'#Store}
	    
	    % set a unique ident for each new connected client
	    idclient <- @idclient+1
	    {Show 'server: setident:'#@idclient}
	    %try {Client setident(@idclient)}
	    %catch
	     %  senderror
	    %then
	     %  skip
	    %end
	 % Add the list of client to new connected client
	 if @clientset\=nil then
	    try  {Client addsetclient(@clientset @identset @identclient @membership)}
	       skip
	    catch senderror then
	       skip
	    end  
	 else
	    skip
	 end
         % broadcast Client to all clients
	 {ForAll @clientset
	  proc {$ Item}
	     try   {Item clientadded(Client @idclient Username#UserId)} %modification to add clientid
	     catch senderror then
		skip
	     end   		     
	  end
	 }	 
	 clientset <- Client|@clientset
	 identset  <- UserId|@identset % added by metwally
	 identclient <- Client#UserId|@identclient %added by metwally
	 membership <- Username#UserId|@membership
	 {Show 'server:client set'#@clientset}
	 {Show 'server:endconnectstore:'#Client#@membership}
	 {Show 'server:endconnectstore:'#@identset#@identclient}	 
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Broadcast new added object to all client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth addobjtostore(Objid State Class Init Client)
	 {Show 'server:addobjtostore'#Objid#State#Class#Init}
	 listobjects <- Objid#Client|@listobjects
	 {ForAll @clientset
	  proc {$ Item}
	     if Item\=Client then 
		try {Item addobj(Objid State Class Init)}
		catch senderror then
		   skip
		end  
	     else
		skip
	     end
	  end
	 }
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      % Rule 15 Operations after Client disconnect 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
      meth disconnectstore(Client)
	 % remove all locks of Client
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	 proc{Getuser Id Li ?User}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  User=H.1
	       else
		  {Getuser Id Tail ?User}
	       end
	    end
	 end
	 Keys Idclient User
      in
	 lock
	    {Show 'server:disconnectclientabnormally:'#Client}
	    {Dictionary.keys @locks ?Keys}
	    {ForAll Keys
	     proc {$ Item}
		if {Dictionary.get @locks Item $} ==Client then
		   {Dictionary.remove @locks Item}
		else
		   skip
		end
	     end
	    }
	    Idclient ={Getident Client @identclient $}
	    User ={Getuser Idclient @membership $}
	    clientset <- {List.subtract @clientset Client $}
	    identset <- {List.subtract @identset Idclient $}
	    identclient <-{List.subtract @identclient Client#Idclient $}
	    membership <- {List.subtract @membership User#Idclient $}
	    {Show 'server:disconnectclientabnormally:end'#Client#Idclient#@identset#@identclient}
	 end
      end
      
      meth normaldisconnectstore(Client ?Sync)
	 % remove all locks of Client
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	 proc{Getuser Id Li ?User}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  User=H.1
	       else
		  {Getuser Id Tail ?User}
	       end
	    end
	 end
	  proc{Getprocwatcher Cref Listproc ?Proc}
	    case Listproc of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Proc=H.2
	       else
		  {Getprocwatcher Cref Tail ?Proc}
	       end
	    end
	 end
	 Keys Idclient User ProcWatch
      in
	 lock
	    {Show 'server:disconnectclientnormally'#Client}
	    {Dictionary.keys @locks ?Keys}
	    {ForAll Keys
	     proc {$ Item}
		if {Dictionary.get @locks Item $} ==Client then
		   {Dictionary.remove @locks Item}
		else
		   skip
		end
	     end
	    }
	    Idclient ={Getident Client @identclient $}
	    User ={Getuser Idclient @membership $}
	    clientset <- {List.subtract @clientset Client $}
	    identset <- {List.subtract @identset Idclient $}
	    identclient <-{List.subtract @identclient Client#Idclient $}
	    membership <- {List.subtract @membership User#Idclient $}
	    {Show 'server:disconnectclientnormally:end'#Client#Idclient#@identset#@identclient}
	 end
	 %ProcWatch
	 
	 ProcWatch={Getprocwatcher Client @listofwatcher $}
	 try
	    {Client deinstallwatcher(ProcWatch Fault)}
	    {Show 'server:watcherdeinstallfromclient:'#Idclient}
	    %{Client installtempwatcher(ProcT Fault)}
	 catch E
	 then
	    {Show cannotdeinstalwatcherfromnormaldisconnectedclientserver#Idclient#E}
	 end
	 {ForAll @clientset
	  proc {$ Items}
	     try   {Items clientdisconnected(Client Idclient User)} %modification to disconnec client
		
	     catch senderror then
		{Show adjustclienterror}   %skip
	     end
	     
	  end
	 }
	 
	 Sync=unit
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      % Build lock table attribute at server after election
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth buildlocks(Store)
	 lock
	    {Show 'server:buildlocks:'#Store}
	    {ForAll Store
	     proc {$ X }
		Objid Client in
		if X\=nil then
		   X=Objid#_#Client
		   {Dictionary.put @locks Objid Client}
		else
		   skip
		end
	     end
	    }
	 end
      end       
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Build clientset at server after selection
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth setclientset(Listclient IdentList IdentClient Membership Clientref Serv) 
	 fun{Maximum L}
	    case L of nil then 0
	    []H1|T1 then
	       {Max2 H1 T1}
	    end
	  end
	 fun{Max2 M L}
	    case L of nil then M
	    []H|T then
	       {Max2  {Value.max M H $} T}
	    end
	 end
	
      in 
	 
	 clientset <- Listclient
	 %idclient  <- MaxId
	 identset <- IdentList
	 identclient <- IdentClient
	 membership <- Membership
	 idobject <- 100 % to enhance
	 idclient <- {Maximum @identset}
         % Install watcher on on clients
	 {ForAll @clientset
	  proc{$ C}
	     proc{Proc E FS}
		{self procwatcher(C)}
	     end
	  in
	     try
		{C installwatcher([permFail] Proc Fault)}			      
		{Show 'server:watcherinstalledon:'#C}
		procwatcherclient <- Proc
		listofwatcher <- C#Proc|@listofwatcher
	     catch cannotinstallwatcher
	     then
		{Show installwatchererror}
	     end
	  end
	 }
	 
	 server <- Serv
	 %procwatcherclient <- Proc
	 % sent to clientref to do set notify recovery
	 %try
	 %   {Clientref recovery}
	 %catch senderror
	 %then
	 %   {Show recoveryerror}
	 %end
	 {Show 'server: setclientset:'#@clientset#@identset#@idclient#@identclient}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% ask if object broadcasted
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth ask(N ?Result)
        % ask all clients if N present
	 
	try  
	   {ForAll @clientset
	    proc {$ C}
	       X in
	       try
		  {C askpresent(N ?X)}
		  {Wait X}
	       catch senderror then
		  skip
	       end
	       if X==true then
		  raise found end
	       else
		  skip
	       end
	    end
	   }
	   
	   Result=false
	catch found
	then Result=true
	end
	
      end      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Clients Synchronous kill
      %%%%%%%%%%%%%%%%%%%%%%%%%%
      meth killstore()
	 {ForAll @clientset
	  proc{$ C}
	     End in
	     try
		{C kill(End)}		      
		{Wait End}
	     catch senderror
	     then
		skip
	     end
	  end
	 }
	 %{Application.exit 0} 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%
      % Server kill
      %%%%%%%%%%%%%%%%%%%%%%%%%%
      meth killserver()
	 {Application.exit 0}
	 %{OS.system "kill -9"#" "#{OS.getPID} _}
      end
  
     %  meth disconnect(Client Id User)
% 	 Keys Meni 
%       in
% 	 lock
% 	    {Show 'normal:disconnectclient:'#Client}
% 	    {Dictionary.keys @locks ?Keys}
% 	    {ForAll Keys
% 	     proc {$ Item}
% 		if {Dictionary.get @locks Item $} ==Client then
% 		   {Dictionary.remove @locks Item}
% 		else
% 		   skip
% 		end
% 	     end
% 	    }
% 	    Meni={List.nth @identset {List.length @identset} $}
	    
% 	    clientset <- {List.subtract @clientset Client $}
% 	    identset <- {List.subtract @identset Id $}
% 	    identclient <-{List.subtract @identclient Client#Id $}
% 	    membership <- {List.subtract @membership User#Id $}
% 	    {ForAll @clientset
% 	     proc {$ Items}
% 		try   {Items clientdisconnected(Client Id User)} %modification to disconnec client
% 		catch senderror then
% 		   {Show adjustclienterror}
% 		   skip
% 		end   		     
% 	     end
% 	    }
% 	  % if this disconnected client was the leader then tell everybody to create a new server
	    
% 	    if Meni == Id then
% 	       {Show 'the leader is disconnecting began a new selection'}
% 	       {ForAll @clientset
% 		proc {$ Items}
% 		   try   {Items deinstalloldwatcher} %modification to disconnec client
% 		   catch senderror then
% 		      {Show 'fail deinstalloldwatcher'}
% 		      skip
% 		   end   		     
% 		end
% 	       }
% 	       {ForAll @clientset
% 		proc {$ Items}
% 		   try   {Items watcheronserver(permfail)} %modification to disconnecclient
% 		   catch senderror then
% 		      {Show 'fail to do new leader selection'}
% 		      skip
% 		   end   		     
% 		end
% 	       }
% 	    else
% 	       skip
% 	    end
% 	    % deinstall my self from the newleader
% 	 %clientset <- {List.subtract @clientset Client $}
% 	 end
%       end
   
      %%%%%%%%%%%%%%%%%%%%%%%%%%
      % Detect termination
      %%%%%%%%%%%%%%%%%%%%%%%%%%
      meth termination(X)
	 lock
	    %{System.show @terminate}
	    terminate <- @terminate +1
	    if @terminate==100 then
	       try X=0
	       catch _ then
		  skip
	       end
	       terminate <- 0
	    else
	       {Delay 200}
	       {@server termination(X)}	      
	    end
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% method to test if server alive
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth ping()
	 skip
      end
      
      meth getuniqueident(Id)
	 idobject <-  @idobject +1
	 Id=@idobject
      end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % leaving method
      %%%%%%%%%%%%%%%%%%%%%%%%%

      meth leave(Client ?Sync)
	 % remove all locks of Client
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	 proc{Getuser Id Li ?User}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  User=H.1
	       else
		  {Getuser Id Tail ?User}
	       end
	    end
	 end
	  proc{Getprocwatcher Cref Listproc ?Proc}
	    case Listproc of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Proc=H.2
	       else
		  {Getprocwatcher Cref Tail ?Proc}
	       end
	    end
	 end
	 Keys Idclient User ProcWatch
      in
	 lock
	    {Show 'server:leave'#Client}
	    {Dictionary.keys @locks ?Keys}
	    {ForAll Keys
	     proc {$ Item}
		if {Dictionary.get @locks Item $} ==Client then
		   {Dictionary.remove @locks Item}
		else
		   skip
		end
	     end
	    }
	    Idclient ={Getident Client @identclient $}
	    User ={Getuser Idclient @membership $}
	    clientset <- {List.subtract @clientset Client $}
	    identset <- {List.subtract @identset Idclient $}
	    identclient <-{List.subtract @identclient Client#Idclient $}
	    membership <- {List.subtract @membership User#Idclient $}
	    {Show 'server:leave:end'#Client#Idclient#@identset#@identclient}
	 end
	 %ProcWatch
	 
	 ProcWatch={Getprocwatcher Client @listofwatcher $}
	 try
	    {Client deinstallwatcher(ProcWatch Fault)}
	    {Show 'server:watcherdeinstallfromleavingclient:'#Idclient}
	    %{Client installtempwatcher(ProcT Fault)}
	 catch E
	 then
	    {Show cannotdeinstalwatcherfromnormaldisconnectedclientserver#Idclient#E}
	 end
	 {ForAll @clientset
	  proc {$ Items}
	     try   {Items clientdisconnected(Client Idclient User)} %modification to disconnec client
		
	     catch senderror then
		{Show adjustclienterror}   %skip
	     end
	     
	  end
	 }
	 
	 Sync=unit
      end

      
      meth newinformation(NewRef Client Idclient Username)
	 % we must update all membership information too 
	 idclient <- Idclient
	 clientset <- Client|@clientset
	 identset  <- Idclient|@identset % added by metwally
	 identclient <- Client#Idclient|@identclient %added by metwally
	 membership <- Username#Idclient|@membership
	 {ForAll NewRef
	  proc{$ Objid}
	     listobjects <- Objid#Client|@listobjects
	  end
	 }
	 % if the server have other client as the case of permfial then
	 % the new server should propagate this to all of them also!!!
      end
      meth normalwork
	 requestlock <- true
      end

      meth mergestore(Client Username ClientId OldStore ProcMerge)
	 lock
	    Lastcommitransid
	    proc {ProcT E FS}
	       {Show 'temporary fail detected'}
	    end
	    proc{Proc E FS}
	       {self procwatcher(Client)}
	    end
	    proc{Getidref Id Li ?Ref}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  Ref=H.1
	       else
		  {Getidref Id Tail ?Ref}
	       end
	    end
	 end
	    Leader Meni
	 in
	    {Show 'server:initmergestore:'#Client}
	    % set  the server attribute
	    try
	       {Client installwatcher([permFail] Proc Fault)}
	       {Show 'serveratmerge:watcherinstalleokclient:'#Client#@clientset#@identclient#@membership}
	       {Client installtempwatcher(ProcT Fault)}
	       procwatcherclient <- Proc
	       listofwatcher <- Client#Proc|@listofwatcher
	    catch cannotinstallwatcher
	    then
	       {Show installwatchererror}
	    end
	    {ForAll @clientset
	     proc {$ Item}
		%now the client will pause automatically because the sever did lock_all
		% now evey client will receive no_lock during remerge operation
		try   {Item clientadded(Client ClientId Username#ClientId)} 
		catch senderror then
		   skip
		end   		     
	     end
	    }
	     % get leader
	    {Show menimumid#Meni}
	    Meni={List.nth @identset {List.length @identset} $}
	    {Show menimumid#Meni}
	    if Meni >  ClientId then
	       Leader <- Client
	    else
	       Leader={Getidref Meni @identclient $}   
	    end
	    %{List.nth @clientset 1 Leader} % this is not good
	    
	    try	      
    	      % Store and Lastcommitransid will be bound at client send this to the leader
	       {Leader remregestore(OldStore ProcMerge)} % leader must sign all object as locked objects
	    catch senderror then skip
	    end
	    
	    if @clientset\=nil then
	       
	       try  {Client mergeaddsetclient(@clientset @identset @identclient @membership)} 
		  skip
	       catch senderror then
		  skip
	       end  
	    else
	       skip
	    end


	    clientset <- Client|@clientset
	    identset  <- ClientId|@identset % added by metwally
	    identclient <- Client#ClientId|@identclient %added by metwally
	    membership <- Username#ClientId|@membership
	    {Show 'serverafremerge:client set'#@clientset}
	    {Show 'server:endmergestore:'#Client#@membership}
	    {Show 'server:endmergestore:'#@identset#@identclient}	

	    requestlock <- false % send no request for any trans until remerege is completed
             % broadcast Client to all clients
	    % {ForAll @clientset
% 	     proc {$ Item}
% 		%now the client will pause automatically because the sever did lock_all
% 		% now evey client will receive no_lock during remerge operation
% 		try   {Item clientadded(Client ClientId Username#ClientId)} 
% 		catch senderror then
% 		   skip
% 		end   		     
% 	     end
% 	    }
	    
	 % Add the list of client to new connected client
	    
	    
	    % clientset <- Client|@clientset
% 	    identset  <- ClientId|@identset % added by metwally
% 	    identclient <- Client#ClientId|@identclient %added by metwally
% 	    membership <- Username#ClientId|@membership
% 	    {Show 'serverafremerge:client set'#@clientset}
% 	    {Show 'server:endmergestore:'#Client#@membership}
 	    {Show 'server:endmergestore:'}	 
	 end 
      end
      % Client Calls

      meth serverpropagatethis(ObjInfo Client)
	 {Show 'server propagate this start'}
	 {ForAll @clientset
	  proc {$ Item}
	     
	     if Item\=Client then 
		try	      
		   {Item propagatethis(ObjInfo)} %
		catch senderror  then
		   {Show 'this client is disconnected'}
		end
	     else skip 
	     end
	  end
	  
	 }
	 {Show 'server propagate newinformation'}
      end

      %%%%%%%%% leave merge pototcol %%%%%%%%%%%%
      
     meth request_locks_for_delete(Objid Client Lock Synch)
	% Lock = Yes or No at the end Synch=unit
	lock
	   
	    terminate <- 0
	    {Show 'server:request_locks_for_delete:begin:'}
	    %try
	   if {Dictionary.condGet @locks Objid noentry $}==noentry
	   then
	      Lock=yes
	      {Dictionary.put @locks Objid Client}
	   else  
	      Lock=no
	   end
	    % Lock = yes
	   {Show 'server:request_locks_for_delete:end:'}
	   if Lock==yes
	   then
	      {Show 'server: i will send to all clients to delete this obj'}
	      %for all clients send delete this object
	      {ForAll @clientset
	       proc {$ Item}
		  
		  if Item\=Client then 
		     try	      
			
		     % send delete thid obj from your store
			{Show 'server send to this client delete object'}
			{Item delete_object(Objid)} %
			
		      %{Wait Synch}
		      % delete this object from server
			%listobjects <- {List.subtract @listobjects Objs#Client $}
			%{Dictionary.remove @locks Objid}
		     catch
			senderror
		     then
			skip
		     end
		  else
		     skip
		  end
	       end
	      }
	      listobjects <- {List.subtract @listobjects Objid#Client $}
	      {Dictionary.remove @locks Objid}
	      {Show 'server a delete all the information about this obj'}
	   else
	      {Show 'server: no lock for delete'}
	      skip
	   end
	   Synch=unit   
	end
     end
   end
end




 
